-- ScriptContextTooltipTemplates.lua
-- Core tooltip template functions for ContextTooltip add-on.
-- 11:55 17.11.2008
--------------------------------------------------------------------------------
-- CONSTANTS
Global( "MAX_LEVEL_DIFFERENCE", 4 )
Global( "MIN_SPELL_RANGE", 5 )
Global( "WEAPON_SPEED_MODIFIER", 10 )
Global( "sysItemQualityStyle", {
	[ ITEM_QUALITY_JUNK ] = "Junk",
	[ ITEM_QUALITY_GOODS ] = "Goods",
	[ ITEM_QUALITY_COMMON ] = "Common",
	[ ITEM_QUALITY_UNCOMMON ] = "Uncommon",
	[ ITEM_QUALITY_RARE ] = "Rare",
	[ ITEM_QUALITY_EPIC ] = "Epic"
} )
Global( "sysBonusStyle", {
	[ -1 ] = "StatDebuffed", 
	[ 0 ] = "StatNormal", 
	[ 1 ] = "StatBuffed"
} )
Global( "sysCharHeaderStyle", {
	dead = "Dead", 
	friendly = "Friendly",
	neutral = "Neutral",
	aggressive = "Aggressive"
} )
Global( "sysInnateStats", {
	[ INNATE_STAT_STRENGTH ] = "strength", --
	[ INNATE_STAT_MIGHT ] = "might", --
	[ INNATE_STAT_DEXTERITY ] = "dexterity", --
	[ INNATE_STAT_AGILITY ] = "agility", --
	[ INNATE_STAT_STAMINA ] = "stamina", -- 
	[ INNATE_STAT_PRECISION ] = "precision", -- 
	[ INNATE_STAT_HARDINESS ] = "hardiness", --
	[ INNATE_STAT_INTELLECT ] = "intellect", --
	[ INNATE_STAT_INTUITION ] = "intuition", --
	[ INNATE_STAT_SPIRIT ] = "spirit", --
	[ INNATE_STAT_WILL ] = "will", --
	[ INNATE_STAT_RESOLVE ] = "resolve", --
	[ INNATE_STAT_WISDOM ] = "wisdom", --
	[ INNATE_STAT_LETHALITY ] = "lethality" --
} )
Global( "sysResistance", {
	[ RESISTANCE_ELEMENTAL ] = "elemental", --
	[ RESISTANCE_DIVINE ] = "divine", --
	[ RESISTANCE_NATURE ] = "nature" --
} )
Global( "sysReputationClass", {
	[ REPUTATION_LEVEL_HOSTILITY ] = "RepHostility",
	[ REPUTATION_LEVEL_ENEMY ] = "RepEnemy",
	[ REPUTATION_LEVEL_UNFRIENDLY ] = "RepUnfriendly",
	[ REPUTATION_LEVEL_NEUTRAL ] = "RepNeutral",
	[ REPUTATION_LEVEL_KINDLY ] = "RepKindly",
	[ REPUTATION_LEVEL_FRIENDLY ] = "RepFriendly",
	[ REPUTATION_LEVEL_CONFIDENTIAL ] = "RepConfidential"
} )
Global( "sysReputationColor", {
	[ REPUTATION_LEVEL_HOSTILITY ] = { a = 255, r = 255 * 0.75, g = 0, b = 0 },
	[ REPUTATION_LEVEL_ENEMY ] = { a = 255, r = 255 * 0.75, g = 0.33 * 255 * 0.75, b = 0 },
	[ REPUTATION_LEVEL_UNFRIENDLY ] = { a = 255, r = 255 * 0.75, g = 0.67 * 255 * 0.75, b = 0 },
	[ REPUTATION_LEVEL_NEUTRAL ] = { a = 255, r = 255 * 0.75, g = 255 * 0.75, b = 0 },
	[ REPUTATION_LEVEL_KINDLY ] = { a = 255, r = 0.67 * 255 * 0.75, g = 255 * 0.75, b = 0 },
	[ REPUTATION_LEVEL_FRIENDLY ] = { a = 255, r = 0.33 * 255, g = 255 * 0.75, b = 0 },
	[ REPUTATION_LEVEL_CONFIDENTIAL ] = { a = 255, r = 0, g = 255 * 0.75, b = 0 }
} )
Global( "sysIndent", {
	[ 0 ] = "Indent0",
	[ 1 ] = "Indent1",
	[ 2 ] = "Indent2",
	[ 3 ] = "Indent3",
	[ 4 ] = "Indent4"
} )
Global( "HP_COLOR_AGGRESSIVE", { r = 255; g = 12; b = 12; a = 255 } ) --red
Global( "HP_COLOR_NEUTRAL", { r = 200; g = 200; b = 12; a = 255 } ) -- yellow
Global( "HP_COLOR_FRIENDLY", { r = 36; g = 134; b = 12; a = 255 } ) -- green
Global( "HP_COLOR_DEAD", { r = 64; g = 64; b = 64; a = 255 } ) -- grey
Global( "COLOR_MANA", { r = 12; g = 74; b = 136; a = 255 } ) -- blue
--------------------------------------------------------------------------------
-- DIGGING DATA FOR TEMPLATE
--------------------------------------------------------------------------------
function GetDurationValuedTextComponent( sysFormat, value )
	value = value or 0
	
	if value > 0 then
		local valuedText = common.CreateValuedText()
		valuedText:SetFormat( GetText( sysFormat ) )
		valuedText:SetVal( "value", common.FormatInt( value, "%d" ) )
		return valuedText
	end
end
--------------------------------------------------------------------------------
function GetDurationTextValues( sysFormat, class, duration )
	duration = duration or 0
	
	if math.floor( duration / 1000 ) > 0 then
		local count = SplitTime( duration )
	
		return {
			format = GetText( sysFormat ),
			style = class,
			hours = GetDurationValuedTextComponent( "DurationHours", math.floor( count.hours ) ),
			minutes = GetDurationValuedTextComponent( "DurationMinutes", math.floor( count.minutes ) ),
			seconds = GetDurationValuedTextComponent( "DurationSeconds", math.floor( count.seconds ) ),
			--milliseconds = GetDurationValuedTextComponent( "DurationMilliseconds", count.milliseconds )
		}
	end
end
--------------------------------------------------------------------------------
function IsPrimaryStat( statId )
	local class = unit.GetClass( avatar.GetId() )
	
	if statId == INNATE_STAT_PRECISION then
		return true
	end
	
	if class.manaType == MANA_TYPE_ENERGY then
		if statId == INNATE_STAT_STRENGTH
		or statId == INNATE_STAT_MIGHT
		or statId == INNATE_STAT_DEXTERITY then
			return true
		end
	elseif class.manaType == MANA_TYPE_MANA then
		if statId == INNATE_STAT_INTELLECT
		or statId == INNATE_STAT_INTUITION
		or statId == INNATE_STAT_SPIRIT then
			return true
		elseif ( class.className == "NECROMANCER" or class.className == "PRIEST" ) and statId == INNATE_STAT_WISDOM then
			return true
		end
	end

	return false
end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
templates[ TOOLTIP_NONE ].GetDataFrom = function( self, params )
	return nil
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SIMPLE ].GetDataFrom = function( self, params )
	local data = {}
	data.name = GetNameFromInfo( params, nil )
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SIMPLEDESC ].GetDataFrom = function( self, params )
	local data = {}
	data.name = GetNameFromInfo( params, nil )
	data.desc = GetDescFromInfo( params, nil )
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_MAPMARK ].GetDataFrom = function( self, params )
	local data = nil
	
	local avatarId = avatar.GetId()
	--trace( "avatarId = ", avatarId )
	
	if not avatarId then
		return data
	end
	
	data = {}
	local unitName = object.GetName( avatarId )
	if not IsEWs( unitName ) then
		data.name = { value = unitName, format = GetText( "Header" ), style = "tip_green" }
	end
	
	local zone = cartographer.GetCurrentZone()
	data.zone = { value = zone.zoneName }
	
	if IsValidWs( zone.subZoneName ) then
		data.subZone = { value = zone.subZoneName }
	end
	
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_UNIT ].GetDataFrom = function( self, params )
	local data = nil

	local avatarId = avatar.GetId()
	local avatarLevel = 1
	if avatarId then
		avatarLevel = unit.GetLevel( avatarId )
	end
	
	data = {}
	data.unitId = params.unitId
	
	local isPlayer = unit.IsPlayer( params.unitId )--returns: boolean
	local isAbleToAttack = unit.IsAbleToAttack( params.unitId )--returns: boolean
	local isAggressive = unit.IsAggressive( params.unitId )--returns: boolean
	local isDead = unit.IsDead( params.unitId )--returns: boolean
	local isEnemy = unit.IsEnemy( params.unitId )--returns: boolean
	local isFriend = unit.IsFriend( params.unitId )--returns: boolean
	local isTrainer = unit.IsTrainer( params.unitId )--returns: boolean
	local isVendor = object.IsVendor( params.unitId )--returns: boolean
	local isUsable = unit.IsUsable( params.unitId )--returns: boolean
	--local Quality = unit.GetQuality( params.unitId ) --UNIT_QUALITY_COMMON, UNIT_QUALITY_ELITE, UNIT_QUALITY_FLAVOR_ELITE, UNIT_QUALITY_MINI_BOSS
	--local TrainerClass = unit.GetTrainerClass( params.unitId ) --className: string, manaType: number, name: WString
	--local canUse = unit.CanUse( params.unitId ) --boolean
	
	local unitFaction = unit.GetFaction( params.unitId )--returns: isSame: boolean, isFriend: boolean, isPassive: boolean, name: WString
	if not common.IsEmptyWString( unitFaction.name ) then
		data.faction = {}
		data.faction.value = unitFaction.name
		data.faction.style = "tip_white"
	end
	
	local name = {
		format = GetText( "Header" ),
		value = ValidateWs( params.name or object.GetName( params.unitId ) )
	}
	
	local health = unit.GetHealth( params.unitId )
	local mana = unit.GetMana( params.unitId )
	
	data.health = { state = health.percents }
	data.mana = { state = mana.percents, color = COLOR_MANA }
	
	if isPlayer then
		if isFriend and not isEnemy then
			name.style = sysCharHeaderStyle.friendly
			data.health.color = HP_COLOR_FRIENDLY
		elseif not isFriend and isEnemy then
			name.style = sysCharHeaderStyle.aggressive
			data.health.color = HP_COLOR_AGGRESSIVE
		else
			name.style = sysCharHeaderStyle.neutral
			data.health.color = HP_COLOR_NEUTRAL
		end
	else
		local factionLevel = unit.GetReputationLevel( params.unitId )
		name.style = sysReputationClass[ factionLevel ]
		data.health.color = sysReputationColor[ factionLevel ]
	end

	if isDead then
		name.style = sysCharHeaderStyle.dead
		data.health.color = HP_COLOR_DEAD
	end
	
	data.name = name
		
	local unitLevel = unit.GetLevel( params.unitId )	--returns: number
	data.level = { format = GetText( "CharacterLevel" ) }
	
	if not isFriend
	and ( unitLevel - avatarLevel ) > MAX_LEVEL_DIFFERENCE then
		data.level.value = GetText( "UnknownLevel" )
		
	else
		data.level.value = common.FormatInt( unitLevel, "%d" )
	end
	
	local unitClass = unit.GetClass( params.unitId )	--returns: table: className, manaType, name: WString
	--trace( "unitClass = ", unitClass )
	if unitClass then
		data.class = { format = GetText( "CharacterClass" ), value = ValidateWs( unitClass.raceClassName ) }
	end
	
	local unitRace = unit.GetRace( params.unitId ) --returns: table: sysCreatureRace: string, characterRace: WString
	if unitRace then
		data.race = { value = GetTextFrom( "Race", unitRace.sysCreatureRace ), format = GetText( "CharacterRace" ) }
	end
	
	local unitSex = nil
	if isPlayer then
		unitSex = unit.GetSex( params.unitId )--returns: sex: number, name: WString, raceSexName: WString
	end
	
	if unitSex and not common.IsEmptyWString( unitSex.raceSexName ) then
		data.sex = { value = unitSex.raceSexName, format = GetText( "CharacterSex" ) }
	end
	
	data.creature = { format = GetText( "CharacterInfo" ) }
	
	if isPlayer then
		data.creature.value = data.sex.value
		data.creature.value2 = data.class.value
		data.creature.value3 = data.level.value
	else
		data.creature.value = data.race.value
		data.creature.value3 = data.level.value
	end

	local unitTitle = unit.GetTitle( params.unitId )
	if IsValidWs( unitTitle ) then
		data.desc = { style = "tip_golden", value = unitTitle, multiline = true }
	end
	
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ITEM ].GetDataFrom = function( self, params )
	local data = nil
	
	local itemInfo = avatar.GetItemInfo( params.itemId )
	--trace( "itemInfo = ", tabletolog( itemInfo ) )
	
	if not itemInfo then
		return data
	end
	
	local avatarId = avatar.GetId()
	local avatarLevel = 1
	if avatarId then
		avatarLevel = unit.GetLevel( avatarId )
	end
	
	local tryEquip = avatar.GetItemDressConditions( params.itemId )
	--trace( "tryEquip = ", tryEquip )
	--------------------------------------
	data = {}
	data.itemId = params.itemId
	data.name = GetNameFromInfo( params, itemInfo )
	data.desc = GetDescFromInfo( params, itemInfo )
	
	data.name.style = itemInfo.quality and sysItemQualityStyle[ itemInfo.quality ]
	--------------------------------------
	if itemInfo.isQuestOperator then
		data.questOperator = { format = GetText( "QuestStarter" ) }
	end
	--------------------------------------
	if itemInfo.isQuestRelated then
		data.questRelated = { format = GetText( "QuestRelated" ) }
	end
	--------------------------------------
	if itemInfo.isGuildCreator then
		data.guildCreator = { format = GetText( "GuildCreator" ) }
	end
	--------------------------------------
	if itemInfo.dressSlot ~= DRESS_SLOT_UNDRESSABLE then
		data.dress = { value = GetDressSlotName( itemInfo.dressSlot ), format = GetText( "Default" ) }
		data.class = { value = itemInfo.className, format = GetText( "DefaultRight" ) }

		if tryEquip.sysFirstCondition ~= "ENUM_DressResult_Success"
		and ( tryEquip.failedConditions[ "ENUM_DressResult_WrongCharacterClass" ]
		or tryEquip.failedConditions[ "ENUM_DressResult_NotProficient" ] ) then
			data.class.style = sysCharHeaderStyle.aggressive
		end
	end
	--------------------------------------
	local misc = itemInfo.bonus.misc

	if itemInfo.isWeapon then
		data.speed = { 
			value = common.FormatFloat( WEAPON_SPEED_MODIFIER / misc.weaponSpeed, "%.1f" ),
			format = GetText( "WeaponSpeed" )
		}
		data.damage = {
			value = common.FormatInt( misc.minDamage, "%d" ),
			value2 = common.FormatInt( misc.maxDamage, "%d" ),
			format = GetText( "DamageMinMax" )
		}
		
		local dps = (misc.minDamage + misc.maxDamage) / ( 2 * misc.weaponSpeed )
		data.dps = { value = common.FormatFloat( dps, "%.2f" ), format = GetText( "DamagePerSecond" ) }

		data.armorPiercing = {
			format = itemInfo.isArmorPiercing and GetText( "ArmourPiercing" ) or GetText( "DefaultRight" )
		}
	end
	
	if misc.spellPower and misc.spellPower ~= 0 then
		data.spellPower = { format = GetText( "SpellPower" ), value = common.FormatInt( misc.spellPower, "%d" ), style = "tip_white" }
	end
	--------------------------------------
	if misc.armor ~= 0 then
		local bonus = misc.armorBonus
		data.armor = { value = common.FormatInt( misc.armor, "%d" ), format = GetText( "Armor" ), style = bonus and bonus > 0 and "StatBuffed" or nil }
	end
	--------------------------------------
	local innateStats = itemInfo.bonus.innateStats
	
	for id, entry in innateStats do
		local sysId = sysInnateStats[ id ]
		local value = entry.base
		
		local class = unit.GetClass( avatar.GetId() )
		local specialStat = IsPrimaryStat( id )
		
		if value ~= 0 then
			data[ sysId ] = {
				stat = GetTextFrom( "InnateStats", entry.sysName ),
				value = common.FormatInt( value, "%+d" ),
				format = GetText( "InnateStatsBonus" ),
				style = sysBonusStyle[ Sign( value ) ],
				background = specialStat and {
					color = { a = 0.75, r = 0, g = 0.5, b = 0 }
				} or nil
			}
			data[ sysId .. "Icon" ] = { icon = specialStat and GetTexture( "StatStar" ) or GetTexture( "Empty" ), size = 20 }
		end
	end
	--------------------------------------
	local resistances = itemInfo.bonus.resistances
	
	for id, entry in resistances do
		local sysId = sysResistance[ id ]
		local value = entry.base
		
		if value ~= 0 then
			data[ sysId ] = {
				stat = GetTextFrom( "Resistances", entry.sysName ),
				value = common.FormatInt( value, "%+d" ),
				format = GetText( "ResistanceBonus" ),
				style = sysBonusStyle[ Sign( value ) ]
			}
			data[ sysId .. "Icon" ] = { icon = GetTexture( sysId ), size = 20 }
		end
	end
	--------------------------------------
	if itemInfo.spellId then
		local spellInfo = avatar.GetSpellInfo( itemInfo.spellId )
		
		if spellInfo then
			data.effect = {
				format = GetText( "Use" ),
				value = spellInfo.description,
				style = spellInfo.canRunAvatar and "tip_green" or "tip_red",
				miltiline = true
			}
			data.effectForeIcon = { icon = spellInfo.texture, size = 34 }
		end
	end
	--------------------------------------
	if itemInfo.craftingSkillIds and itemInfo.craftingSkillIds[ 0 ] then
		data.craftSkill = {}
		
		for id = 0, table.getn( itemInfo.craftingSkillIds ) do
			local skillInfo = avatar.GetSkillInfo( itemInfo.craftingSkillIds[ id ] )
			data.craftSkill[ id ] = { format = GetText( "CraftingSkill" ), value = ValidateWs( skillInfo.name ), style = "body" }
		end
	end
	--------------------------------------
	if GetTableSize( itemInfo.craftingComponents ) > 0 then
		data.craftComponents = {}

    for i = 0, GetTableSize( itemInfo.craftingComponents ) - 1 do
			local componentInfo = {}
			if itemInfo.craftingComponents[ i ] then
				componentInfo = avatar.GetComponentInfo( itemInfo.craftingComponents[ i ] )
			end

			data.craftComponents[ i + 1	] = componentInfo.image or GetTexture( PlaceholderOrange )
		end
	end
	--------------------------------------
	if itemInfo.requiredLevel > 1
	and ( itemInfo.dressSlot ~= DRESS_SLOT_UNDRESSABLE or itemInfo.spellId or itemInfo.isUsable )
	then
		data.requiredLevel = { format = GetText( "ItemRequiredLevel" ), value = common.FormatInt( itemInfo.requiredLevel, "%d" ), style = "body" }
		
		if tryEquip.sysFirstCondition ~= "ENUM_DressResult_Success"
		and tryEquip.failedConditions[ "ENUM_DressResult_WrongCreatureLevel" ]
		or itemInfo.requiredLevel > avatarLevel then
			data.requiredLevel.style = sysCharHeaderStyle.aggressive
		end
	end
	--------------------------------------
	if itemInfo.runeInstrumentInfo then
		data.runeInstrument = { style = "tip_green" }
		
		if itemInfo.runeInstrumentInfo.isCombiner then
			data.runeInstrument.value = GetText( "RuneInstrumentCombiner" )
			
		elseif itemInfo.runeInstrumentInfo.isExtractor then
			data.runeInstrument.value = GetText( "RuneInstrumentExtractor" )
		end
	end
	--------------------------------------
	local insertedRuneInfo = itemInfo.runeId and avatar.GetItemInfo( itemInfo.runeId )
	
	if itemInfo.runeInfo then
		local runeInfo = itemInfo.runeInfo
		if runeInfo then
			data.runeName = {
				format = GetText( "RuneZodiacName" ),
				name = ValidateTextValue( runeInfo.zodiacSignName ),
				desc = ValidateTextValue( runeInfo.zodiacSignDescription ),
				multiline = true,
				style = "tip_white"
			}
		
			data.runeLevel = { value = common.FormatInt( runeInfo.runeLevel, "%d" ), format = GetText( "RuneLevel" ), style = "tip_green" }
			data.runeForeIcon = { icon = itemInfo.icon or GetTexture( "PlaceholderOrange" ), size = 44 }
		end

	elseif insertedRuneInfo then
		local runeInfo = insertedRuneInfo.runeInfo
		if runeInfo then
			data.runeInfo = {
				count = runeInfo.runeLevel or 1,
				icon = insertedRuneInfo.icon,
				rune = ValidateTextValue( runeInfo.zodiacSignName ),
				level = common.FormatInt( runeInfo.runeLevel, "%d" ),
				style = "tip_green"
			}
			
			data.offensiveBonus = runeInfo.offensiveBonus > 0 and {
				format = GetText( "RuneOffensiveBonus" ),
				value = common.FormatFloat( runeInfo.offensiveBonus, "%f3K5" ),
				style = "tip_green"
			} or nil
			data.defensiveBonus = runeInfo.defensiveBonus > 0 and {
				format = GetText( "RuneDefensiveBonus" ),
				value = common.FormatFloat( runeInfo.defensiveBonus, "%f3K5" ),
				style = "tip_green"
			} or nil
		end
		
	elseif itemInfo.canInsertRune == true then
		data.canInsertRune = { format = GetText( "RuneCanInsert" ), style = "tip_green" }
		data.canInsertRuneIcon = { icon = GetTexture( "RuneLevelGold" ), size = 26 }
	end
	--------------------------------------
	if itemInfo.isBoxKey then
		data.isBoxKey = { value = GetText( "IsBoxKey" ), style = "tip_green" }
	end
	--------------------------------------
	if itemInfo.boxInfo then
		data.isBoxLocked = itemInfo.boxInfo.isBoxLocked
		and { value = GetText( "BoxLocked" ), style = "tip_white" }
		or { value = GetText( "BoxUnLocked" ), style = "tip_green" }
	end
	--------------------------------------
	if itemInfo.foragingInfo then
		local info = itemInfo.foragingInfo
		local skillInfo = avatar.GetSkillInfo( info.skillId )
		
		if skillInfo then
			data.foraging = { format = GetText( "Foraging" ), value2 = ValidateWs( skillInfo.name ), style = "body" }
			data.foraging.value1 = info.isInstrument and GetText( "ForagingInstrument" ) or GetText( "ForagingItem" )
		end
	end
	--------------------------------------
	if itemInfo.disassemblerInfo then
		local info = itemInfo.disassemblerInfo
		data.disassembler = { format = GetText( "Disassembler" ) }
	end
	--------------------------------------
	if itemInfo.isBound then
		data.binding = { format = GetText( "Soulbound" ), style = "tip_golden" }
		
	elseif itemInfo.binding ~= ITEM_BINDING_BIND_NONE then
		data.binding = { value = GetTextFrom( "ItemBinding", sysItemBindingId[ itemInfo.binding ] ), style = "tip_golden" }
	end
	--------------------------------------
	local vendorId = avatar.GetInterlocutor()
	local isVendor = avatar.IsInteractorVendor()
	--trace( "vendorId = ", vendorId, ", isVendor = ", isVendor )
	
	local sysFactionName = vendorId and object.IsUnit( vendorId ) and unit.GetFaction( vendorId ).sysName
	--trace( "sysFactionName = ", sysFactionName )
	
	local reputationInfo = sysFactionName and avatar.GetReputationInfo( sysFactionName )
	--trace( "reputationInfo = ", reputationInfo )
	
	local reputation = reputationInfo and reputationInfo.level or REPUTATION_LEVEL_HOSTILITY
	--trace( "reputation = ", sysReputationLevelId[ reputation ] )
	
	if params.tooltip == TOOLTIP_ITEM_VENDOR
	and isVendor
	and itemInfo.requiredReputationLevel ~= 0 then
		data.requiredReputation = {
			format = GetText( "RequiredReputationLevel" ),
			value = GetTextFrom( "ReputationLevel", sysReputationLevelId[ itemInfo.requiredReputationLevel ] )
		}

		data.requiredReputation.style = itemInfo.requiredReputationLevel > reputation and "Aggressive" or "Friendly"
	end
	--------------------------------------
	local price = isVendor and ( params.tooltip == TOOLTIP_ITEM and itemInfo.sellPrice or itemInfo.buyPrice ) or itemInfo.basePrice
	local count = price * math.max( 1, itemInfo.stackCount )
	
	data.money = count < 0
		and { format = GetText( "Default" ), value = GetText( "NoSellPrice" ), style = "tip_grey" }
		or { count = count }
	--------------------------------------
	if itemInfo.requiredHonorRank ~= 0 then
		local honor = unit.GetHonorInfo( avatarId ).rank
		--trace( "honor = ", sysHonorRankId[ honor ] )
		
		data.requiredHonor = {
			format = GetText( "RequiredHonorRank" ),
			value = GetTextFrom( "HonorRank", sysHonorRankId[ itemInfo.requiredHonorRank ] )
		}
		
		if itemInfo.requiredHonorRank > honor then
			data.requiredHonor.style = "Aggressive"
			
		else
			data.requiredHonor.style = "Friendly"
		end
	end
	--------------------------------------
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ITEM_VENDOR ].GetDataFrom  = templates[ TOOLTIP_ITEM ].GetDataFrom
--------------------------------------------------------------------------------
Global( "requrementsValues", { 
	{ name = "number", GetVal = function( number ) return common.FormatInt( number, "%dK5" ) end },
	{ name = "name", GetVal = function( name ) return name end },
	{ name = "dress", GetVal = function( dressSlot ) return GetDressSlotName( dressSlot ) end },
	{ name = "itemClass", GetVal = function( itemClass ) return itemClass end },
	{ name = "sysRace", GetVal = function( sysRace ) return GetTextFrom( "Race", sysRace ) end }, 
	{ name = "sysPsionicContactType", GetVal = function( sysPsionicContactType ) return GetTextFrom( "PsionicContact", sysPsionicContactType ) end },
	{ name = "reputationLevel", GetVal = function( reputationLevel ) return GetTextFrom( "ReputationLevel", sysReputationLevelId[ reputationLevel ] ) end }
})
--------------------------------------------------------------------------------
  -- AFCT_Or -  ""   
  -- AFCT_And -  
  -- AFCT_Not -  
--------------------------------------------------------------------------------
function SetRequrementsValues( valuedText, req )
	for i, val in requrementsValues do
		local strName = val.name
		local reqValue = req[ strName ]
		if reqValue then
			if reqValue then
				local resultText = val.GetVal( reqValue )
				if resultText then
					valuedText:SetVal( strName, resultText )
				end
			end
		end
	end
end
--------------------------------------------------------------------------------
function AddRequirementHeader( data, indent, text )
	local valuedText = common.CreateValuedText()
	valuedText:SetFormat( GetText( "RequirementHeader" ) )
	valuedText:SetVal( "indent", GetText( sysIndent[ indent ] ) )
	valuedText:SetVal( "header", text ) 
	table.insert( data, { format = GetText( "Default" ), value = valuedText, style = "tip_white" } )
end
--------------------------------------------------------------------------------
function AddParentRequirementInParent( data, indent, parentType )
  -- AFCT_Or -  ""   
	if parentType == AFCT_Or then
		AddRequirementHeader( data, indent, GetText( "AFCT_Or" ) )
	end
end
--------------------------------------------------------------------------------
function AddParentRequirementHeader( data, indent, parentType, currentType )
  -- AFCT_Not -  
	if parentType == AFCT_Not then

		if currentType == AFCT_Or then
			AddRequirementHeader( data, indent, GetText( "AFCT_Or_Not" ) )
		elseif currentType == AFCT_And then
			AddRequirementHeader( data, indent, GetText( "AFCT_And_Not" ) )
		end

		return
	end

  -- AFCT_Or -  ""   
	if currentType ~= AFCT_Not then
		AddParentRequirementInParent( data, indent, parentType )
	end

	if currentType == AFCT_Or then
		AddRequirementHeader( data, indent, GetText( "AFCT_Or_Main" ) )
	end
end
--------------------------------------------------------------------------------
function GetSimpleRequirementText( parentType, sysCause )
	if parentType == AFCT_Not then
	  -- AFCT_Not -  
		return GetTextFrom( "ActionRequirementsNot", sysCause )
	else
		return GetTextFrom( "ActionRequirements", sysCause )
	end
end
--------------------------------------------------------------------------------
function FillSpellRequirement( req, data, indent, parentType, prevReqInParentType )
	if req.type then
		-- parent (Or/And/Not)
		local newIndent = indent 
		if req.type ~= AFCT_Not then
			newIndent = indent + 1
		end

		AddParentRequirementHeader( data, indent, parentType, req.type )
		FillSpellRequirementsTable( req.childs, data, newIndent, req.type )

	else
		if prevReqInParentType == AFCT_Or or prevReqInParentType == AFCT_And then
			AddParentRequirementInParent( data, indent, parentType )
			indent = indent + 1
		end

		-- simple		
		local textStyle = nil
		if not req.success then
			textStyle = "tip_red"
		end

		local valuedTextCause = common.CreateValuedText()
		valuedTextCause:SetFormat( GetSimpleRequirementText( parentType, req.sysCause ) )
		SetRequrementsValues( valuedTextCause, req )

		local valuedText = common.CreateValuedText()
		valuedText:SetFormat( GetText( "RequirementLine" ) )
		valuedText:SetVal( "cause", valuedTextCause )
		valuedText:SetVal( "indent", GetText( sysIndent[ indent ] ) )

		table.insert( data, { format = GetText( "Default" ), value = valuedText, style = textStyle } )
	end
end
--------------------------------------------------------------------------------
function FillSpellRequirementsTable( reqTable, data, indent, parentType )
	if reqTable then
		local size = GetTableSize( reqTable )
		for i = 0, size - 1 do
			local prevReqInParentType = ( reqTable[ i - 1 ] and reqTable[ i - 1 ].type )
			FillSpellRequirement( reqTable[ i ], data, indent, parentType, prevReqInParentType )
		end
	end
end
--------------------------------------------------------------------------------
function FillSpellRequirementsBaseTable( reqTable, data, tableName, showHeader )
	if reqTable and GetTableSize( reqTable ) > 0 then
		if showHeader then
			table.insert( data, { format = GetText( "Default" ), value = GetText( tableName ), style = "tip_blue" } )
		end
		FillSpellRequirementsTable( reqTable, data, 0, AFCT_And )

		return true
	end

	return false
end
--------------------------------------------------------------------------------
function FillSpellRequirementsData( spellRequirements, data )
	local hasTable = FillSpellRequirementsBaseTable( spellRequirements.casterConditions, data, "RequirementCasterConditions", true  )	
	FillSpellRequirementsBaseTable( spellRequirements.casterResources, data, "RequirementCasterResources", not hasTable )	

	hasTable = FillSpellRequirementsBaseTable( spellRequirements.targetConditions, data, "RequirementTargetConditions", true )	
	FillSpellRequirementsBaseTable( spellRequirements.targetResources, data, "RequirementTargetResources", not hasTable )	
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SPELL ].GetDataFrom = function( self, params )
	local data = nil
	
	if not params.spellId then
		return data
	end
	
	local spellInfo = avatar.GetSpellInfo( params.spellId )
	--trace( "spellInfo = ", spellInfo )
	
	if not spellInfo then
		return data
	end

	data = {}
	data.spellId = params.spellId
	data.name = GetNameFromInfo( params, spellInfo )
	data.desc = GetDescFromInfo( params, spellInfo )
	
	data.rank = { format = GetText( "SpellRank" ), value = common.FormatInt( spellInfo.rank, "%d" ), style = "tip_white" }
	data.level = { format = GetText( "SpellRequiredLevel" ), value = common.FormatInt( spellInfo.level, "%d" ), style = "tip_white" }
	
	local current = spellInfo.currentValues
	spellInfo.prepareDuration = current and current.prepareDuration or spellInfo.prepareDuration
	spellInfo.range = current and current.range or spellInfo.range
	spellInfo.radius = current and current.radius or spellInfo.radius
	spellInfo.manaCost = current and current.manaCost
	spellInfo.predictedCooldown = current and current.predictedCooldown

	spellInfo.requirements = avatar.GetSpellRequirements( params.spellId )
	
	data.castDuration = {
		format = spellInfo.prepareDuration == 0 and GetText( "SpellCastImmediate" ) or GetText( "SpellCastDuration" ),
		value =  spellInfo.prepareDuration ~= 0 and common.FormatFloat( spellInfo.prepareDuration / 1000, "%g") or nil,
		style = "tip_white"
	}
	
	data.element = { format = GetText( "SpellSubElement" ), value = common.GetAddonRelatedGroupText( "SubElement", spellInfo.sysSubElement ), style = "tip_white" }
	
	if not spellInfo.launchWhenReady then
		data.launchWhenReady = { value = GetText( "SpellPreCast" ), style = "tip_green" }
	end
	
	if spellInfo.targetType ~= SPELL_TYPE_SELF then
		data.range = {
			format = GetText( "SpellRange" ),
			style = "tip_white",
			value = spellInfo.range > MIN_SPELL_RANGE
				and common.FormatFloat( spellInfo.range, "%g" )
				or GetText( "SpellRangeClose" )
		}
	end
	
	if spellInfo.radius > 0 then
		data.radius = { format = GetText( "SpellRadius" ), value = common.FormatFloat( spellInfo.radius, "%g" ), style = "tip_white" }
	end
	
	if spellInfo.cooldownRemainingMs > 0 then
		data.cooldownRemains = {
			format = GetText( "SpellCooldown" ),
			value = common.FormatFloat( Round( spellInfo.cooldownRemainingMs / 1000 ), "%g" ),
			style = "Aggressive"
		}
	elseif spellInfo.predictedCooldown then
		data.cooldownDuration = {
			format = spellInfo.predictedCooldown ~= 0 and GetText( "SpellCooldownDuration" ) or GetText( "SpellCooldownImmediate" ),
			value = spellInfo.predictedCooldown ~= 0 and common.FormatFloat( Round( spellInfo.predictedCooldown / 1000 ), "%g" ) or nil,
			style = "tip_blue"
		}
	end
	
	data.manaCost = {
		format = unit.GetClass( avatar.GetId() ).manaType == MANA_TYPE_MANA and GetText( "SpellManaCost" ) or GetText( "SpellEnergyCost" ),
		value = spellInfo.manaCost ~= 0 and common.FormatInt( spellInfo.manaCost, "%d" ) or GetText( "SpellCostFree" ),
		style = "tip_golden"
	}

	if spellInfo.requirements then
		data.requirements = {}
		FillSpellRequirementsData( spellInfo.requirements, data.requirements )
	end

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_BUFF ].GetDataFrom = function( self, params )
	local data
	
	local buff = unit.GetBuff( params.unitId, params.index )
	if buff == nil then
		return data
	end

	data = { unitId = params.unitId, index = params.index }
	data.name = GetNameFromInfo( params, buff )
	data.desc = GetDescFromInfo( params, buff )
	data.duration = GetDurationTextValues( "BuffDuration", "tip_blue", buff.remainingMs )

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ABILITY ].GetDataFrom = function( self, params )
	local data = nil
	
	if not params.abilityId then
		return data
	end
	
	local abilityInfo = avatar.GetAbilityInfo( params.abilityId )
	--trace( "abilityInfo = ", abilityInfo )
	if not abilityInfo then
		return data
	end
	
	data = {}
	data.name = GetNameFromInfo( params, abilityInfo )
	data.desc = GetDescFromInfo( params, abilityInfo )
	
	data.rank = { format = GetText( "SpellRank" ), value = common.FormatInt( abilityInfo.rank, "%d" ), style = "body" }
	
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_COMPONENT ].GetDataFrom = function( self, params )
	local data = nil
	
	if not params.componentId then
		return data
	end
	
	local componentInfo = avatar.GetComponentInfo( params.componentId )
	--trace( "componentInfo = ", componentInfo )
	if not componentInfo then
		return data
	end
	data = { name = GetNameFromInfo( params, componentInfo ), desc = GetDescFromInfo( params, componentInfo ) }

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_RECIPE ].GetDataFrom = function( self, params )
	local data = nil
	
	if not params.recipeId then
		return data
	end
	
	local recipeInfo = avatar.GetRecipeInfo( params.recipeId )
	--trace( "recipeInfo = ", recipeInfo )
	if not recipeInfo then
		return data
	end
	
	data = {}
	data.name = GetNameFromInfo( params, recipeInfo )
	data.desc = GetDescFromInfo( params, recipeInfo )

	data.craftComponents = {}
		
	for i = 0, 4 do
		local componentInfo = nil

		if recipeInfo.components[ i ] then
		  if params.armorCraft then
				componentInfo = avatar.GetItemInfo( recipeInfo.components[ i ] )
				
			else
				componentInfo = avatar.GetComponentInfo( recipeInfo.components[ i ] )
			end
			
			if componentInfo then
				if componentInfo.icon then
					data.craftComponents[ i + 1	] = componentInfo.icon
					
				elseif componentInfo.image then
					data.craftComponents[ i + 1	] = componentInfo.image
					
				else
					data.craftComponents[ i + 1	] = GetTexture( "PlaceholderOrange" )
				end
			end
		end
	end

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_USABLE_DEVICE ].GetDataFrom = templates[ TOOLTIP_SIMPLE ].GetDataFrom
--------------------------------------------------------------------------------
templates[ TOOLTIP_ASTRAL_UNIT ].GetDataFrom = function( self, params )
	local data = nil
	
	if not params.unitId then
		return data
	end
	
	data = { unitId = params.unitId }
	
	local name = object.GetName( params.unitId )
	data.name = { format = GetText( "Header" ), value = params.name or ValidateWs( name ), style = sysCharHeaderStyle.aggressive }

	local health = astralUnit.GetHealthLevel( params.unitId )
	data.health = { state = health, color = HP_COLOR_AGGRESSIVE }
	if health <= 0 then
		data.health.color = HP_COLOR_DEAD
	end
	
	local level = astralUnit.GetLevel( params.unitId )
	data.level = { format = GetText( "CharacterLevel" ), value = common.FormatInt( level, "%d" ) }
	
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_FAIRY ].GetDataFrom = function( self, params )
	local data = nil
	local fairyInfo = unit.GetFairyInfo( params.unitId )
	--trace( "fairyInfo = ", tabletolog( fairyInfo ) )
	
	if not fairyInfo then
		return data
	end

	data = { unitId = params.unitId }
	data.name = GetNameFromInfo( params, fairyInfo )
	data.name.style = "Header"
	data.desc = GetDescFromInfo( params, fairyInfo )
	data.level = { format = GetText( "FairyLevel" ),value = common.FormatInt( fairyInfo.level, "%d" ) }
	
	if fairyInfo.isHungry then
		data.isHungry = { value = GetText( "FairyIsHungry" ), style = "Aggressive" }
	end
	
	data.remainingMs = fairyInfo.remainingMs 
	and {
		format = GetText( "FairyRemainingMs" ),
		value = common.FormatInt( Round( fairyInfo.remainingMs / 1000 ), "%d" )
	}
	
	local avatarClass = unit.GetClass( avatar.GetId() )
	
	data.healthBonus = { format = GetText( "FairyHealthBonus" ), value = common.FormatInt( fairyInfo.healthBonus, "%+d" ), style = "Friendly" }
	data.manaBonus = {
		format = GetText( avatarClass.manaType == MANA_TYPE_MANA and "FairyManaBonus" or "FairyEnergyBonus" ),
		value = common.FormatInt( fairyInfo.manaBonus, "%+d" ),
		style = avatarClass.manaType == MANA_TYPE_MANA and "tip_blue" or "tip_golden"
	}

	local fairyZodiacSignInfo = unit.GetFairyZodiacSignInfo( params.unitId )
	--trace( "fairyZodiacSignInfo = ", tabletolog( fairyZodiacSignInfo ) )
	
	data.fairyName = {
		format = GetText( "FairyZodiacName" ),
		name = ValidateTextValue( fairyZodiacSignInfo.name ),
		desc = ValidateTextValue( fairyZodiacSignInfo.description ),
		multiline = true,
		style = "tip_white"
	}
	data.fairyForeIcon = { icon = fairyZodiacSignInfo.image or GetTexture( "PlaceholderOrange" ), size = 44 }
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_TALENT ].GetDataFrom = function( self, params )
	local data
	local talentInfo
	
	if params.field then
		talentInfo = avatar.GetFieldTalentInfo( params.field, params.row, params.column )
		
	else
		talentInfo = avatar.GetBaseTalentInfo( params.row, params.column )
	end
	--trace( "talentInfo = ", tabletolog( talentInfo ) )
	
	if not talentInfo then
		return data
	end
	
	local spentTP = avatar.GetSpentTalentPoints()
	--trace( "spentTP = ", spentTP )
	local pointsId = avatar.GetCurrencyId( "TalentPoint" )
	--trace( "pointsId = ", pointsId )
	local pointsInfo = pointsId and avatar.GetCurrencyInfo( pointsId )
	--trace( "pointsInfo = ", pointsInfo )
	
	local rubyId = avatar.GetCurrencyId( "Ruby" )
	--trace( "rubyId = ", rubyId )
	local rubyInfo = rubyId and avatar.GetCurrencyInfo( rubyId )
	--trace( "rubyInfo = ", rubyInfo )
	
	data = { field = params.field, row = params.row, column = params.column }
	
	if talentInfo.current then
		local current = talentInfo.current
		
		if current.spellId then
			data.current = templates[ TOOLTIP_SPELL ]:GetDataFrom( { spellId = current.spellId } )
			
		elseif current.abilityId then
			data.current = templates[ TOOLTIP_ABILITY ]:GetDataFrom( { abilityId = current.abilityId } )
		end
		
		if data.current then
			data.current.rank.format = GetText( "TalentRankCurrent" )
		end
	end
	
	if talentInfo.next then
		local next = talentInfo.next
		
		if next.spellId then
			data.next = templates[ TOOLTIP_SPELL ]:GetDataFrom( { spellId = next.spellId } )
			
		elseif next.abilityId then
			data.next = templates[ TOOLTIP_ABILITY ]:GetDataFrom( { abilityId = next.abilityId } )
		end
		
		if data.next then
			data.next.rank.format = GetText( "TalentRankNext" )
		end
		
		if params.field then
			if rubyInfo then
				data.nextCost = {
					format = GetText( "TalentCost" ),
					value = common.FormatInt( 1, "%d" ),
					currency = ValidateWs( rubyInfo.name ),
					style = rubyInfo.value < 1 and "Aggressive" or "Friendly"
				}
				
				if not talentInfo.canUpdate
				and rubyInfo.value >= 1
				and not talentInfo.isLearned then
					data.canUpdate = {
						value = GetText( "TalentCantUpdateField" ),
						style = "Aggressive"
					}
				end
			end
			
		else
			if pointsInfo then
				data.nextCost = {
					format = GetText( "TalentCost" ),
					value = common.FormatInt( next.rank, "%d" ),
					currency = ValidateWs( pointsInfo.name ),
					style = pointsInfo.value < next.rank and "Aggressive" or "Friendly"
				}
				
				if spentTP < talentInfo.requiredSpentTP then
					data.canUpdate = {
						format = GetText( "TalentCantUpdateBase" ),
						value = common.FormatInt( talentInfo.requiredSpentTP - spentTP, "%d" ),
						currency = ValidateWs( pointsInfo.name ),
						style = "Aggressive"
					}
				end
			end
		end
	end
		
	if talentInfo.isEmpty then
		data.name = { format = GetText( "Header" ), value = GetText( "TalentEmptyField" ) }
	end

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_REPUTATION ].GetDataFrom = function( self, params )
	local data
	local repInfo = avatar.GetReputationInfo( params.faction )

	if not repInfo then
		return data
	end

	data = { faction = params.faction }
	data.label = { format = GetText( "Header" ), value = GetText( "ReputationLabel" ) }
	data.name = { value = ValidateWs( params.name or repInfo.factionName ), style = "tip_white" }
	data.level = {
		format = GetText( "ReputationLevel" ),
		level = GetTextFrom( "ReputationLevel", sysReputationLevelId[ repInfo.level ] ),
		style = sysReputationClass[ repInfo.level ],
		rep = common.FormatInt( repInfo.rep, "%d" ),
		repNext = ( repInfo.repNextLevel == -1 or repInfo.repNextLevel == nil ) and GetText( "ReputationInfinity" ) or common.FormatInt( repInfo.repNextLevel, "%d" )
	}
	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_HEARTHSTONE ].GetDataFrom = function( self, params )
	local data

	if not avatar.HasHearthStone() then
		return data
	end

	local cooldown = avatar.GetHearthStoneCooldown() -- remainingMs, durationMs: number( int )
	--trace( "cooldown = ", cooldown )
	
	local locatorInfo = avatar.GetHearthStoneLocator() -- map, zone, locator: WString
	--trace( "locatorInfo = ", locatorInfo )
	
	data = { inactive = cooldown.remainingMs > 0 or nil }
	data.name = { format = GetText( "Header" ), value = params.name and ValidateWs( params.name ) or GetText( "HearthstoneName" ) }
	data.desc = {
		format = GetText( "Default" ),
		style = "tip_golden",
		value = params.desc and ValidateWs( params.desc ) or 
		( data.inactive and GetText( "HearthstoneDescInactive" ) or GetText( "HearthstoneDescActive" ) )
	}
	if cooldown.remainingMs > 0 then
		data.remainingMs = {
			format = GetText( "HearthstoneCooldown" ),
			value = common.FormatFloat( math.floor( cooldown.remainingMs / 1000 ), "%g" ),
			style = "Aggressive"
		}
	end
	data.map = { format = GetText( "HearthstoneMap" ), map = ValidateWs( locatorInfo.map ), zone = ValidateWs( locatorInfo.zone ), style = "tip_white" }
	--data.locator = { format = GetText( "HearthstoneLocator" ), value = ValidateWs( locatorInfo.locator ), style = "tip_white" }

	return data
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_INNATE_STAT ].GetDataFrom = function( self, params )
	local data

	if not params.statId then
		return data
	end

	local stats = avatar.GetInnateStats()
	local info = stats[ params.statId ]
	
	if not info then
		return data
	end
	
	data = { statId = params.statId }
	
	local delta = info and info.effective - info.base or 0
	local style = delta > 0 and "StatBuffed" or delta < 0 and "StatBuffed" or "StatNormal"
		
	local derivativeSuffix = "_N"
	
	if info.sysName == "ENUM_IS_Might" then
		derivativeSuffix = info.secondary.current.N == 0 and "_K" or "_N"
		
	elseif info.sysName == "ENUM_IS_Spirit" then
		local class = unit.GetClass( avatar.GetId() )
		local isManaUser = class and class.manaType == MANA_TYPE_MANA or false
		derivativeSuffix = isManaUser and "_NK" or "_N"
		
	elseif info.sysName == "ENUM_IS_Will"
	or info.sysName == "ENUM_IS_Resolve" then
		derivativeSuffix = "_NK"
	end
	
	data.name = {
		format = GetText( "Header" ),
		style = "tip_white",
		value = GetTextFrom( "InnateStats", info.sysName )
	
	}
	data.base = { format = GetText( "StatBase" ), value = common.FormatInt( info.base, "%d" ), style = "StatNormal" }
	
	data.talents = info.talents ~= 0
		and {
			format = GetText( "StatTalents" ),
			value = common.FormatInt( info.talents, "%+d" ),
			style = info.talents < 0 and "StatDebuffed" or "StatBuffed"
		}
		or nil
	
	data.equipment = info.equipment ~= 0 and {
			format = GetText( "StatEquipment" ),
			value = common.FormatInt( info.equipment, "%+d" ),
			style = info.equipment < 0 and "StatDebuffed" or "StatBuffed"
		}
		or nil
		
	data.buffs = info.buffs ~= 0 and {
		format = GetText( "StatBuffs" ),
			value = common.FormatInt( info.buffs, "%+d" ),
			style = info.buffs < 0 and "StatDebuffed" or "StatBuffed"
		}
		or nil
		
	data.current = {
		format = GetText( "Stat_" .. info.sysName .. derivativeSuffix ),
		valueN = common.FormatFloat( info.secondary.current.N, "%f3K5" ),
		valueK = common.FormatFloat( info.secondary.current.K, "%f3K5" ),
		style = "tip_golden"
	}
	
	data.next = {
		format = GetText( "Stat_" .. info.sysName .. derivativeSuffix ),
		valueN = common.FormatFloat( info.secondary.next.N, "%f3K5" ),
		valueK = common.FormatFloat( info.secondary.next.K, "%f3K5" ),
		style = "tip_golden"
	}
		
	data.statnext = {	format = GetText( "StatNext" ), value = common.FormatInt( info.effective, "%d" ),
		style = ( info.effective + 1 - info.base ) > 0 and "StatBuffed" or delta < 0 and "StatBuffed" or "StatNormal"
	}
	
	data.important = IsPrimaryStat( params.statId ) and {
		format = GetText( "Stat_Important" ),
		style = "tip_green"
	}
	or nil

	return data
end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- PREPARING TEMPLATE
--------------------------------------------------------------------------------
templates[ TOOLTIP_NONE ].AssembleContentFrom = function ( self, data )
	return {}
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SIMPLE ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name }
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SIMPLEDESC ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc }
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_MAPMARK ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.zone },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.subZone }
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_UNIT ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.creature },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.faction },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = TOOLTIP_CONTENT.PROGRESS_BAR, data.health,
			update = {
				[ "EVENT_UNIT_HEALTH_PERCENTAGE_CHANGED" ] = function( params )
					local health
					if params.unitId == data.unitId then
						health = unit.GetHealthPercentage( params.unitId )
						health = {
							state = health,
							color = unit.IsDead( params.unitId )
							and HP_COLOR_DEAD
							or sysReputationColor[ unit.GetReputationLevel( params.unitId ) ]
						}
					end
					return health
				end
			}
		},
		--[[{ type = TOOLTIP_CONTENT.PROGRESS_BAR, data.mana,
			update = {
				[ "EVENT_UNIT_MANA_PERCENTAGE_CHANGED" ] = function( params )
					local mana
					if params.unitId == data.unitId then
						mana = unit.GetManaPercentage( params.unitId )
						mana = { state = mana, color = COLOR_MANA }
					end
					return mana
				end
			}
		}--]]
		update = {
			[ "EVENT_REPUTATION_LEVEL_CHANGED" ] = function( event )
				return true
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ITEM ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.binding },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.isBoxKey },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.questOperator },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.questRelated },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.guildCreator },
		{ type = TOOLTIP_CONTENT.RUNE_INFO_LINE, data.runeInfo },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.offensiveBonus },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.defensiveBonus },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.runeLevel },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.runeName, data.runeForeIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.canInsertRune, data.canInsertRuneIcon },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.runeInstrument },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.dress, data.class },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.damage, data.dps },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.speed, data.armorPiercing },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.spellPower },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.armor, data.armorIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.strength, data.strengthIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.might, data.mightIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.dexterity, data.dexterityIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.agility, data.agilityIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.stamina, data.staminaIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.precision, data.precisionIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.hardiness, data.hardinessIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.intellect, data.intellectIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.intuition, data.intuitionIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.spirit, data.spiritIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.will, data.willIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.resolve, data.resolveIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.wisdom, data.wisdomIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.lethality, data.lethalityIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.elemental, data.elementalIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.divine, data.divineIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.nature, data.natureIcon },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.effect, data.effectForeIcon },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = ( data.money and data.money.count ) and TOOLTIP_CONTENT.MONEY_COUNT or TOOLTIP_CONTENT.SMART_LINE, data.money },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.isBoxLocked },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.craftSkill and data.craftSkill[ 0 ] },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.craftSkill and data.craftSkill[ 1 ] },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.craftSkill and data.craftSkill[ 2 ] },
		{ type = TOOLTIP_CONTENT.ICON_BAR, data.craftComponents },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.foraging },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.disassembler },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.requiredLevel },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.requiredReputation },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.requiredHonor }
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ITEM_VENDOR ].AssembleContentFrom = templates[ TOOLTIP_ITEM ].AssembleContentFrom
--------------------------------------------------------------------------------
function AppendItem( content, listItem )
	if listItem then
		table.insert( content, { type = TOOLTIP_CONTENT.SMART_LINE, listItem } )
	end
end
--------------------------------------------------------------------------------
function AppendList( content, listTable )
	if listTable then
		local size = GetTableSize( listTable )
		for i = 1, size  do
			AppendItem( content, listTable[ i ] )
		end
	end
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_SPELL ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.rank, data.element },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.range },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.radius },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.manaCost },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.castDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.cooldownDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.cooldownRemains,
			update = {
				[ "EVENT_SECOND_TIMER" ] = function( params )
					return {
						format = GetText( "SpellCooldown" ),
						value = common.FormatFloat( Round( avatar.GetSpellInfo( data.spellId ).cooldownRemainingMs / 1000 ), "%g" ),
						style = "Aggressive"
					}
				end
			}
		},
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.launchWhenReady },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		update = {
			[ "EVENT_SPELLBOOK_ELEMENT_EFFECT" ] = function( params )
				return data.spellId:IsEqual( params.id )
			end
		}
	}

	if GetTableSize( data.requirements ) > 0 then
		table.insert( content, { type = TOOLTIP_CONTENT.SEPARATOR, true } )
	end
	AppendList( content, data.requirements )

	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_BUFF ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.duration },
		update = {
			[ "EVENT_UNIT_BUFFS_ELEMENT_CHANGED" ] = function( event )
				return data.unitId == event.unitId and data.index == event.index
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_ABILITY ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.rank },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc }
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_COMPONENT ].AssembleContentFrom = templates[ TOOLTIP_SIMPLEDESC ].AssembleContentFrom
--------------------------------------------------------------------------------
templates[ TOOLTIP_RECIPE ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc }
	}
	if not data.armorCraft then
		table.insert( content, { type = TOOLTIP_CONTENT.ICON_BAR, data.craftComponents } )
	end
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_USABLE_DEVICE ].AssembleContentFrom = templates[ TOOLTIP_SIMPLE ].AssembleContentFrom
--------------------------------------------------------------------------------
templates[ TOOLTIP_ASTRAL_UNIT ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.level },
		{ type = TOOLTIP_CONTENT.PROGRESS_BAR, data.health,
			update = {
				[ "EVENT_ASTRAL_UNIT_HEALTH_PERCENTAGE_CHANGED" ] = function( params )
					local health
					if params.unitId == data.unitId then
						health = astralUnit.GetHealthLevel( params.unitId )
						health = { state = health, color = health > 0 and HP_COLOR_AGGRESSIVE or HP_COLOR_DEAD }
					end
					return health
				end
			}
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_FAIRY ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.level },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.isHungry },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.remainingMs,
			update = {
				[ "EVENT_SECOND_TIMER" ] = function( params )
					local fairyInfo = unit.GetFairyInfo( data.unitId )
					local remainingMs = fairyInfo and fairyInfo.remainingMs or 0
					return {
						format = GetText( "FairyRemainingMs" ),
						value = common.FormatInt( Round( fairyInfo.remainingMs / 1000 ), "%d" )
					}
				end
			}
		},
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.healthBonus },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.manaBonus },
		{ type = TOOLTIP_CONTENT.ICON_SMART_LINE, data.fairyName, data.fairyForeIcon },
		update = {
			[ "EVENT_UNIT_FAIRY_CHANGED" ] = function( params )
				return params.unitId == data.unitId
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_TALENT ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name or data.current and data.current.name or data.next and data.next.name },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.current and data.current.rank, data.current and data.current.element },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.range },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.radius },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.manaCost },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.castDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.cooldownDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.cooldownRemains,
			update = {
				[ "EVENT_SECOND_TIMER" ] = function( params )
					return {
						format = GetText( "SpellCooldown" ),
						value = common.FormatFloat( Round( avatar.GetSpellInfo( data.current.spellId ).cooldownRemainingMs / 1000 ), "%g" ),
						style = "Aggressive"
					}
				end
			}
		},
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.launchWhenReady },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current and data.current.desc },
		{ type = TOOLTIP_CONTENT.SEPARATOR, data.current and data.next },
		{ type = TOOLTIP_CONTENT.DOUBLE_LINE, data.next and data.next.rank--[[, data.next and data.next.element]] },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.nextCost },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.range },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.radius },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.manaCost },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.castDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.cooldownDuration },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.cooldownRemains },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.launchWhenReady },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next and data.next.desc },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.canUpdate },
		update = {
			[ "EVENT_SPELLBOOK_ELEMENT_EFFECT" ] = function( params )
				return data.current.spellId:IsEqual( params.id )
			end
		},
	}

	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_REPUTATION ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.label },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.level },
		update = {
			[ "EVENT_REPUTATION_LEVEL_CHANGED" ] = function( params )
				return params.sysFactionName == data.faction
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_HEARTHSTONE ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.desc },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.remainingMs, 
			update = {
				[ "EVENT_SECOND_TIMER" ] = function( params )
					return {
						format = GetText( "HearthstoneCooldown" ),
						value = common.FormatFloat( math.floor( avatar.GetHearthStoneCooldown().remainingMs / 1000 ), "%g" ),
						style = "Aggressive"
					}
				end
			}
		},
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.map, 
			update = {
				[ "EVENT_HEARTSTONE_LOCATOR_CHANGED" ] = function( params )
					local locatorInfo = avatar.GetHearthStoneLocator()
					return { format = GetText( "HearthstoneMap" ), map = ValidateWs( locatorInfo.map ), zone = ValidateWs( locatorInfo.zone ) }
				end
			}
		},
		--[[{ type = TOOLTIP_CONTENT.SMART_LINE, data.locator, 
			update = {
				[ "EVENT_HEARTSTONE_LOCATOR_CHANGED" ] = function( params )
					return { format = GetText( "HearthstoneLocator" ), value = ValidateWs( avatar.GetHearthStoneLocator().locator ) }
				end
			}
		},]]
		update = {
			[ "EVENT_HEARTHSTONE_COOLDOWN_STARTED" ] = function( params )
				return true
			end,
			[ "EVENT_HEARTHSTONE_COOLDOWN_FINISHED" ] = function( params )
				return true
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
templates[ TOOLTIP_INNATE_STAT ].AssembleContentFrom = function ( self, data )
	local content = {
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.name },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.base },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.talents },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.equipment },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.buffs },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.current },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.statnext },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.next },
		{ type = TOOLTIP_CONTENT.SMART_LINE, data.important },
		update = {
			[ "EVENT_AVATAR_STATS_CHANGED" ] = function( event )
				return true
			end
		}
	}
	return content
end
--------------------------------------------------------------------------------
